/*++

Copyright (c) 2004 Alexander A. Telyatnikov (Alter)

Module Name:
    DbgPrintLog.cpp

Abstract:

Author:
    Alexander A. Telyatnikov (Alter)

Environment:
    User mode only

Notes:

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Revision History:

--*/

#include <windows.h>
#include <stdio.h>
#include "..\Driver\DbgPrnHk.h"
#include "SvcManLib.h"
#include "Privilege.h"
#include "flusher.h"

#define PROD_VER_MJ  0
#define PROD_VER_MN  1
#define PROD_VER_SUB  ""

PWCHAR LogFile = L"DbgPrint.log";
WCHAR LogFileN[MAX_PATH*2];
WCHAR LogFileSuf[MAX_PATH];

PDbgPrnHk_GetMessages_USER_OUT MsgBuffer = NULL;
ULONG BufferSize = 1024*1024;

ULONG MaxLogSize = 1024*1024*128; //128Mb

HANDLE StdIn = NULL;

void
get_messages()
{
    HANDLE h;
    HANDLE fh;

    ULONG len;
    ULONG i, j;
    WCHAR a;
    PWCHAR NumPos;
    ULONG returned;
    BOOLEAN new_line;
    CHAR DbgPrintBuf[256];
    BOOLEAN esc_key = FALSE;
    BOOLEAN pause_key = FALSE;
    BOOLEAN next_log;
    ULONG key = 0;
    ULONG tmp;
    INPUT_RECORD inp;
    ULONG SleepCount;
    CHAR tab3[] = "\t\t\t";
    ULONG CurLogSize;
    SYSTEMTIME SysTime;
    OSVERSIONINFO OsVer;
    PCHAR PlatformStr;

    printf("DbgPrintLog v%d.%d%s (c) by Alter\nHome site http://www.alter.org.ua\n",
        PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);

    h = CreateFile(NT_DbgPrnHk_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                   FILE_SHARE_READ | FILE_SHARE_WRITE,
                   NULL,
                   OPEN_EXISTING,
                   FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
    if(!h || h == (HANDLE)(-1)) {
        printf("Error: can't open driver\n");
        exit(1);
    }

    wcsncpy(LogFileN, LogFile, MAX_PATH-10);
    LogFileN[MAX_PATH-10] = 0;
    len = wcslen(LogFileN);

    NumPos = &LogFileN[len];
    LogFileSuf[0] = 0;
    for(i=len; i>0; i--) {
        a = LogFileN[i-1];
        if(a == '\\') {
            break;
        }
        if(a == '.') {
            NumPos = &LogFileN[i-1];
            wcscpy(LogFileSuf, NumPos);
            break;
        }
    }

    MsgBuffer = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer) {
        printf("Error: can't allocate message buffer\n");
        exit(4);
    }

    i = 0;

    printf("'Esc' - exit           'Space' - pause/restert\n"
           "'N'   - start new log  'F'     - flush log buffer to disk\n");

    while(!esc_key) {

        if(i>0) {
            swprintf(NumPos, L"%d%s", i, LogFileSuf);
        }
        fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        if(fh && fh != (HANDLE)(-1)) {
            CloseHandle(fh);
            fh = NULL;
            i++;
            if(!i) {
                printf("Error: find unused filename\n");
                exit(2);
            }
            continue;
        }
        fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       CREATE_NEW,
                       FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        if(!fh || fh == (HANDLE)(-1)) {
            printf("Error: can't create log-file\n");
            fh = NULL;
            exit(3);
        }
        new_line = TRUE;
        printf("Writing log to %ws\n", LogFileN);
        CurLogSize = 0;
        len = sprintf(DbgPrintBuf, "Log generated by DbgPrintLog v%d.%d%s\n", PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
        WriteFileBuffer(fh, DbgPrintBuf, len);

        GetSystemTime(&SysTime);
        len = sprintf(DbgPrintBuf, "Timestamp: %d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d\n",
            SysTime.wYear,
            SysTime.wMonth,
            SysTime.wDay,
            SysTime.wHour,
            SysTime.wMinute,
            SysTime.wSecond);
        WriteFileBuffer(fh, DbgPrintBuf, len);

        OsVer.dwOSVersionInfoSize = sizeof(OsVer);
        GetVersionEx(&OsVer);
        switch(OsVer.dwPlatformId) {
        case VER_PLATFORM_WIN32s:
            PlatformStr = "Win";
            break;
        case VER_PLATFORM_WIN32_WINDOWS:
            PlatformStr = "Win 9x";
            break;
        case VER_PLATFORM_WIN32_NT:
            PlatformStr = "NT";
            break;
        default:
            PlatformStr = "Unknown";
        }
        len = sprintf(DbgPrintBuf, "OS Version: %d:%s %d.%d (%d)\n  %s\n",
            OsVer.dwPlatformId,
            PlatformStr,
            OsVer.dwMajorVersion,
            OsVer.dwMinorVersion,
            OsVer.dwBuildNumber,
            OsVer.szCSDVersion);
        WriteFileBuffer(fh, DbgPrintBuf, len);

        next_log = FALSE;
        while(!esc_key && !next_log) {
            if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetMessages,
                                NULL,0,
                                MsgBuffer, BufferSize,
                                &returned,NULL)) {
                break;
            }
            if(pause_key) {
                returned = 0;
            }
            if(!returned) {
                SleepCount++;
                if(SleepCount > 5) {
                    SleepCount = 0;
                    new_line = TRUE;
                    FlushFileBuffered(fh);
                }
                Sleep(1000);
            }
            SleepCount = 0;
            j = 0;
            while(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT)) {
                if(MsgBuffer[j].Length) {
                    if(new_line) {
                        len = sprintf(DbgPrintBuf, "%I64d.%d\t"/* "          " "          " "    "*/,
                            MsgBuffer[j].TimeStamp.QuadPart/10000,
                            (ULONG)(MsgBuffer[j].TimeStamp.QuadPart%10000));
                        WriteFileBuffer(fh, DbgPrintBuf, len/*24*/);
                        CurLogSize+=24;
                        new_line = FALSE;
//                    } else {
//                        WriteFileBuffer(fh, tab3, 3);
                    }
                    len = MsgBuffer[j].Length;
                    if(MsgBuffer[j].Msg[len-1] == '\n') {
                        new_line = TRUE;
                    }
                    WriteFileBuffer(fh, &MsgBuffer[j].Msg[0], len);
                    CurLogSize+=len;
                }
                j++;
                returned -= sizeof(DbgPrnHk_GetMessages_USER_OUT);
            }
            if(j < 256) {
                Sleep(100);
            }
            //Sleep(1);
            while(GetNumberOfConsoleInputEvents(StdIn, &key) && key) {
                key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
                if(key && inp.EventType == KEY_EVENT) {
                    if(inp.Event.KeyEvent.bKeyDown &&
                        inp.Event.KeyEvent.wVirtualScanCode == 0x01 /*VK_ESC*/) {
                        esc_key = TRUE;
                        break;
                    }
                    if(inp.Event.KeyEvent.bKeyDown &&
                        inp.Event.KeyEvent.wVirtualScanCode == 0x39 /*VK_SPACE*/) {
                        pause_key = !pause_key;
                        printf("Log %s\n", !pause_key ? "restarted" : "paused");
                        continue;
                    }
                    if(inp.Event.KeyEvent.bKeyDown &&
                        inp.Event.KeyEvent.wVirtualScanCode == 0x31 /*VK_N*/) {
                        next_log = TRUE;
                        printf("next log\n");
                        break;
                    }
                    if(inp.Event.KeyEvent.bKeyDown &&
                        inp.Event.KeyEvent.wVirtualScanCode == 0x21 /*VK_F*/) {
                        printf("flush log\n");
                        FlushFileBuffered(fh);
                        continue;
                    }
                }
            }
            if(CurLogSize > MaxLogSize) {
                next_log = TRUE;
                printf("max log size limit reached -> next log\n");
            }
        } // end of Get Message loop
        FlushFileBuffered(fh);
        CloseHandle(fh);
        fh = NULL;
    }

    if(fh) {
        CloseHandle(fh);
    }
    CloseHandle(h);
}


void main () {
//    HANDLE h;
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;

    LONG l=0;
    PUCHAR buff=NULL;
    PUCHAR buff_j=NULL;
    PUCHAR buff_i=NULL;
//    LONG tmp;
    LONG i=0, j=0;
    ULONG k;
//    WCHAR a;

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    if(argc >= 2) {
        LogFile = argv[1];
    }

    for(i=0; i<argc; i++) {

    }

    StdIn = GetStdHandle(STD_INPUT_HANDLE);

    if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
        printf("Warning: Insufficient privileges\n");
    }

    k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, ".");
    
    k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
    if(k == 1) {
        // Ok
    } else
    if(k == 0) {
        k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);
    }
    if(k < 0) {
        printf("Error: Can't load or install DbgPrnHk driver\n");
        exit(-1);
    }

    get_messages();

    exit(0);
}
